/*
    This program demonstrates how a "race condition" can come up
    when two threads both manipulate a shared variable. In this
    example, the shared variable is the class variable count.
    A race condition is when the result of a piece of code
    depends on the order in which the processor executes the
    threads.

    If you run this program on a multicore machine, you will see
    a fair number (but not a lot) of race conditions. However, if
    you run this program on a single core machine (or restrict the
    affinity of this program, and all of its threads, to a single
    core), then you will see very few race conditions. This shows
    how rare a race condition can be and why race conditions can be
    so hard to detect in real programs.
*/

public class Race_1v4a
{
   // "global" variables
   public  static volatile boolean runFlag = true; // Try removing the volatile keyword.
   public  static volatile long count  = 0;  // This variable is shared by thread1 and thread2.
   private static long count1 = 0;  // This variable is used only by thread1.
   private static long count2 = 0;  // This variable is used only by thread2.

   public static void main(String[] args)
   {
      Function1 f1 = new Function1();
      Function2 f2 = new Function2();

      // Create two threads.
      System.out.println("Race_1v4a: Starting threads:");
      Thread t1 = new Thread( f1 );
      Thread t2 = new Thread( f2 );

      t1.start();
      t2.start();

      try
      {
         Thread.sleep(2000); // Let the threads run for a while.

         runFlag = false; // Signal the threads to stop.

         t1.join();
         t2.join();

         System.out.println("\nThreads stopped.");
         System.out.printf("count = %,d  count1 = %,d  count2 = %,d\n", count, count1, count2);
      }
      catch (InterruptedException e)
      {
         e.getMessage();
      }
   }

   // Give this thread a "pulse".
   private static class Function1 implements Runnable
   {
      public void run()
      {
         long new_count, old_count;
         while (runFlag)
         {
            // Create a race situation with thread2.
            // Increment a global variable and store the results locally.
            old_count = count++;   // 1. read count, 2. write old_count, 3. inc count
            new_count = count;     // 4. read count, 5. write new_count
            // Check that the local and global variables were consistent with each other.
            if (new_count - old_count != 1)
               System.out.printf("Bingo1: new_count - old_count = %d\n", new_count-old_count);
            count1++;
            if (count1 % 1_000_000 == 0) System.out.print("@"); // A "pulse".
         }
         System.out.println("Thread 1 ending.");
      }
   }

   // Give this thread a "pulse".
   private static class Function2 implements Runnable
   {
      public void run()
      {
         long new_count, old_count;
         while (runFlag)
         {
            // Create a race situation with thread1.
            old_count = count++;   // 1. read count, 2. write old_count, 3. inc count
            new_count = count;     // 4. read count, 5. write new_count
            // Check for the race condition.
            if (new_count - old_count != 1)
               System.out.printf("Bingo2: new_count - old_count = %d\n", new_count-old_count);
            count2++;
            if (count2 % 1_000_000 == 0) System.out.print("#"); // A "pulse".
         }
         System.out.println("Thread 2 ending.");
      }
   }
}

